/*  EXC.PRG
This is a GAUSS program which estimates a Conditionally Heteroskedastic and
Skewed Student T process with no covariates

written by:

Bruce E. Hansen
Department of Economics
Social Science Building
University of Wisconsin
Madison, WI 53706-1393
bhansen@ssc.wisc.edu
http://www.ssc.wisc.edu/~bhansen/

*/
library optmum,pgraph;
graphset;

load y[]=franc.dat;      /* Engle-Bollerslev data */
y = ln(y);
t = rows(y)-1;
y = (y[2:t+1] - y[1:t])*100;
ylag = 0|y[1:t-1];
ylag2 = ylag.^2;

output file = e10.out reset;
output off;

@ Conditional Mean Specification: @

let inc = 0;       /*   1  to include intercept                   */
let gm = 0 ;       /*   1  to include garch-m term (st deviation) */
let ar = 0  ;      /*   1  to include AR lag                     */
let sar = 0;       /*   1  to include squared AR lag             */

@ Conditional Variance @
let archs = 1 ;    /*  arch lags on squared errors             */
ig = 0 ;            /*  1 to force I-GARCH                      */

@ Density Parameters @
fattail = 1;         /*  1  allows for student t density          */
skew    = 1;         /*  1  allows for skewed density             */

let talags  = 1 ;     @  lagged errors                    @
let tslags  = 1 ;     @  lagged squared errors            @
let gtlag = 0;      /*  1 allows for g lag  */

let sklags = 1 ;
let sslags = 1 ;
               /*       lags of variable for skew, 0  if no dependence     */

/*  Initial Parameter Vector */
let b = .037 .20 1.03
        -1.10 -0.54 -0.08
         -0.06 -0.13 -0.10;

ta = maxc(talags); sa = maxc(sklags);
ts = maxc(tslags); ss = maxc(sslags);
arch = maxc(archs);  archn=rows(archs);
if ta == 0; tailn = 0; else; tailn = rows(talags); endif;
if ts == 0; tails = 0; else; tails = rows(tslags); endif;
if sa == 0; skewn = 0; else; skewn = rows(sklags); endif;
if ss == 0; skews = 0; else; skews = rows(sslags); endif;
t = rows(y);
initials = 2;

pmean=inc+gm+ar+sar;
pvar  = 2+archn-ig;
pfat  = fattail+tailn+tails+gtlag;
if pmean .ne 0; cmean = seqa(1,1,pmean); endif;
cvar  = seqa(pmean+1,1,pvar);
cfat  = seqa(pmean+pvar+1,1,pfat);

proc (2) = var(b);
local h,delta,t1,e,omega,alpha,beta,e2,elag,yt,bm,om1;
yt = y;
bm = 1;
if inc == 1; yt = yt - b[bm]; bm = bm+1; endif;
if gm .ne 0; delta = b[bm]; bm=bm+1; else; delta = 0; endif;
if ar == 1; yt = yt - ylag*b[bm]; bm=bm+1; endif;
if sar == 1; yt = yt - ylag2*b[bm]; bm=bm+1;  endif;
omega = ones(t,1)*b[1+pmean];
alpha = b[2+pmean:1+pmean+archn];
if ig .ne 1; beta = b[2+pmean+archn]; else; beta = 1; endif;
 beta = beta - sumc(alpha);
if gm .ne 0;
 h = zeros(t,1); e = h; e2 = h;
 h[1] = omega[1]/(1-beta);
 if h[1] < 0; h[1] = 0; endif;
 e[1] = yt[1] - sqrt(h[1])*delta;
 e2[1] = e[1]^2;
 t1 = 2; do while t1 <= t;
  if t1-arch < 1; elag = rev(e2[1:t1-1])|zeros(arch-t1+1,1);
  else; elag = e2[t1-archs]; endif;
  h[t1] = omega[t1] + h[t1-1]*beta + elag'alpha;
  if h[t1] < 0; h[t1] = 0; endif;
  e[t1] = yt[t1] - sqrt(h[t1])*delta;
  e2[t1] = e[t1]^2;
 t1 = t1+1; endo;
else;
 e = yt;
 e2 = e.^2;
 elag = 0|e2[1:t-1];
 t1=2; do while t1 <= arch;
  elag = elag~(zeros(t1,1)|e2[1:t-t1]);
 t1=t1+1; endo;
 omega = omega + elag*alpha;
 om1 = meanc(e2);
 @  om1 = omega[1]/(1-beta); @
 h = recserar(omega,om1,beta);
endif;
if minc(h) == 0; h = 1; endif;
retp(h,e);  endp;

proc st(b,e);
local z,zq,q1,gs,nu,beta,g1,g2;
if fattail == 0; nu = exp(10); else;
 gs = 1+pmean+pvar;
 nu = b[gs];
 if ta > 0;
   z = 0|e[1:t-1]; zq = z;
   q1=1; do while q1 <= ta; zq = zq~(zeros(q1,1)|z[1:t-q1]); q1=q1+1; endo;
   nu = nu + zq[.,talags]*b[gs+1:gs+tailn];
 endif;
 if ts > 0;
   g1 = gs+tailn;
   z = (0|e[1:t-1]).^2; zq = z;
   q1=1; do while q1 <= ts; zq = zq~(zeros(q1,1)|z[1:t-q1]); q1=q1+1; endo;
   nu = nu + zq[.,tslags] * b[g1+1:g1+tails];
 endif;
 if gtlag == 1;
  beta = b[gs+tailn+tails+1];
  nu = recserar(nu,nu[1],beta);
 endif;
 if ta+ts > 0;
  nu = 27.9/(1+exp(-nu)) + 2.1;
 endif;
endif;
retp(nu); endp;

proc tskew(b,e);
local z,zq,q1,gs,lam,sk,e2;
lam = 0;
if skew .ne 0;
 gs = 1+pmean+pvar+pfat;
 lam = b[gs];
 if (sa > 0) or (ss > 0);
  z = 0|e[1:t-1];
  zq = z;
  q1=1; do while q1 <= sa;
    zq = zq~(zeros(q1,cols(z))|z[1:t-q1,.]);
  q1=q1+1; endo;
  lam = lam + zq[.,sklags]*b[gs+1:gs+skewn];
  z = z.^2;  zq = z;
  q1=1; do while q1 <= ss;
    zq = zq~(zeros(q1,cols(z))|z[1:t-q1,.]);
  q1=q1+1; endo;
  lam = lam + zq[.,sslags]*b[gs+skewn+1:gs+skewn+skews];
 endif;
 if sa+ss > 0;
  lam = 1.8/(1+exp(-lam)) - .9;
 endif;
endif;
retp(lam); endp;

proc likes(b);
local r1,r2,h,e,nu,lam,lnc,c,a,b2,b,z,sgn,z2,rat,f;
{h,e} = var(b);
nu = st(b,e);
lam = tskew(b,e);
if (minc(h) .<= 0);
 f = -100000000000;
else;
 r1 = (nu+1)/2;
 r2 = nu - 2;
 lnc = lnfact(r1-1);
 lnc = lnc - lnfact((nu./2) - 1) - ln(r2*pi)*.5;
 c = exp(lnc);
 a = (r2.*c.*lam*4)./(nu-1);
 b2 = ((lam.^2)*3 - a.^2 + 1)./h;
 b = sqrt(b2);
 z = e + a./b;
 sgn = (z .>= 0) - (z .< 0);
 z2 = (z./(sgn.*lam + 1)).^2;
 rat = z2.*b2./r2 + 1;
 f = lnc + ln(b) - ln(rat).*r1;
 f = f[1+initials:rows(f)];
endif;
retp(-f); endp;

proc likes2(b);
local h,e,f;
{h,e}=var(b);
f = ln(h)*.5 + ln(2*pi)/2 + (e.^2)./(h*2);
f = f[1+initials:rows(f)];
retp(f); endp;

proc like(b); retp(sumc(likes(b))); endp;
proc like2(b); retp(sumc(likes2(b))); endp;

/*  *********************** ESTIMATION ********************** */


{b,lik,g,v} = optmum(&like,b);
lik=like(b);
"Likelihood Has Converged";
"Computing Test Statistics";
gr = gradp(&likes,b);
hs = moment(gr,0);
hes = hessp(&like,b);
if minc(eigrs(hes)) > 0;
 hesf = 0;
 v1 = invpd(hes);
 v = v1*hs*v1;
else;
 hesf = 1;
 v = hs;
endif;
se = sqrt(diag(v));
se1 = sqrt(diag(v1));
{h,e} = var(b);
nu = st(b,e);
lam = tskew(b,e);
z = e./sqrt(h);
zi = z[1+initials:t];
zs = zi.^(1~2~3~4);
mz = meanc(zs);
zs = zs - mz';
ti = t-initials;
sez = sqrt(meanc(zs.^2)/ti);
mcgr = moment(cumsumc(gr),0);
nyblom = sumc(diag(mcgr/hs))/ti;
nybloms = diag(mcgr)./(diag(hs)*ti);
if pmean .ne 0;
 nyblomm = sumc(diag(mcgr[cmean,cmean]/hs[cmean,cmean]))/ti;
endif;
nyblomv = sumc(diag(mcgr[cvar,cvar]/hs[cvar,cvar]))/ti;
if fattail == 1;
nyblomf = sumc(diag(mcgr[cfat,cfat]/hs[cfat,cfat]))/ti;
endif;
bout = b~se1~se~nybloms;


/* ************************ DENSITY ESTIMATION ************************ */

nux = 10; hx = 1; lamx = 0;

proc density(x);
local r1,r2,c,a,b2,b,z,sgn,z2,rat,f;
 r1 = (nux+1)/2;
 r2 = nux - 2;
 c = exp(lnfact(r1-1) - lnfact(nux./2 - 1))/sqrt(r2*pi);
 a = (r2.*c.*lamx*4)./(nux-1);
 b2 = ((lamx.^2)*3 - a.^2 + 1)./hx;
 b = sqrt(b2);
 z = x + a./b;
 sgn = (z .>= 0) - (z .< 0);
 z2 = (z./(sgn*lamx + 1)).^2;
 rat = z2.*b2./r2 + 1;
 f =  (rat.^(-r1))*(b*c);
retp(f); endp;

zu = zeros(ti,1);
i = 1; do while i <= ti;
 if fattail*(ta+ts) > 0;
   nux = nu[initials+i]; else; nux = nu; endif;
 if skew*(sa+ss) > 0; lamx = lam[initials+i]; else; lamx = lam; endif;
 xi = zi[i]|-20;
 zu[i] = intsimp(&density,xi,.001);
i = i + 1; endo;
zu = sortc(zu,1);
fzu = seqa(1,1,ti)/ti;
dzu = abs(zu-fzu);
ks = maxc(dzu)*sqrt(ti);
vm = meanc(dzu.^2)*ti;


bn = 1000;
br = seqa(-5,.01,bn);
fz = zeros(bn,1); fe = fz;

i = 1; do while i<=ti;
 if fattail == 0;  nux = 1000;
 else;
  if (ta==0) and (ts==0);
   nux = nu;   else;   nux = nu[initials+i];   endif;
 endif;
 if skew == 0; lamx = 0; else;
  if (sa==0) and (ss==0); lamx = lam; else; lamx = lam[initials+i]; endif;
 endif;
  hx = 1;              fz = fz + density(br);
  hx = h[initials+i] ; fe = fe + density(br);
i = i + 1; endo;
fz = fz./ti; fe = fe./ti;

/*
 * NKERN.PRC
 */

/*
 * vah 10:29 am, Friday, April 24, 1987
 *     11:52 am, Saturday, April 25, 1987
 * msl  5:00 pm, Tuesday, January 31, 1989 (cosmetic changes)
 *     12:20 pm, Wednesday, February 1, 1989
 *
 * The pdf of the vector x is estimated non-parametrically by the kernel
 * method, using pdf-N(0,1) as the kernel.
 *
 * A window of width h is used.
 *    If h=0, default=npt^(-1/5)*std(x)         }
 *    If h<0, default=npt^(-1/5)*std(x)*abs(h)  } see Singh and Ullah (1986)
 *    If h>0, accept value without modification }
 *
 * The points at which estimates are required are in vector b, which is nptx1.
 * Vector b must be sorted.
 * If x is the name of a file, the variable is read from this file;
 * this allows for length(x) > 8000. The variable 'col' refers to the column
 * in the data set to be read. (If x is a vector, the value of 'col' is
 * irrelevant.)
 */
proc nkern(x,h,b,col);
    local npt,f,c,sx,n,j;
    local fl,cf,nbatch,k,z;

    npt = rows(b);
    f   = zeros(npt,1);

    if type(x)==6;       @ --- "x is vec..." --- @

        if h<=0;
            if h==0; c=1; else; c=-h; endif;
            sx = stdc(x);
            h  = c*(npt^(-1/5))*sx;
        endif;
        n = rows(x);
        j = 1;
        do until j>npt;
            f[j] = sumc(pdfn((b[j]-x)/h))/(n*h);
            j = j+1;
        endo;

    elseif type(x)==13;  @ --- "x is string..." --- @
        open fl = ^x;
        if fl==-1;
            "File " $x " could not be opened. Program terminated...";
            end;
        endif;

        n = rowsf(fl);
        cf= colsf(fl);
        if ( col>cf or col<1);
            "Incorrect input, variable 'col'. Program terminated...";
            end;
        endif;

        nbatch = 8000/cf;
        nbatch = minc(n|nbatch);

        j = 1;
        do until j>npt;
            z = seekr(fl,1);
            k = 1;
            do until eof(fl);
                z = submat(readr(fl,nbatch),0,col);
                if k==1 and h<=0;
                    if h==0; c=1; else; c=-h; endif;
                    sx = stdc(z);
                    h  = c*(npt^(-1/5))*sx;
                endif;
                f[j] = f[j] + sumc(pdfn((b[j]-z)/h))/(n*h);
                k = k+1;
            endo; /* k loop */
            j = j+1;
        endo; /* j loop */

        fl = close(fl);

    else ;               @ --- "x is neither vector nor string" --- @
        "Invalid input for variable 'x'. Program terminated...";
        end;
    endif;

retp(f);
endp;


kernz = nkern(z,-.65,br,1);
kerne = nkern(e,-.4,br,1);


output on;
"Conditional Mean: ";
" Intercept        " inc;
" Garch-M          " gm;
" AR lag           " ar;
" Squared          " sar;
"Conditional Variance:";
" Arch lags        " archs';
" I-Garch          " ig;
"Conditional Density:";
" Student T        " fattail;
if fattail == 1;
" Lags on residual " talags';
" Lags on squares  " tslags';
endif;
" Skew             " skew;
if skew == 1;
" Lags on residual " sklags';
" Lags on squares  " sslags';
endif;
"";
"Estimates:";
"";
if hesf == 1;
"  WARNING:  Hessian not positive definite "; "" ;
endif;
if pmean .ne 0;
"Conditional Mean";
bout[1:pmean,.];
"";
endif;
"Conditional Variance";
bout[pmean+1:pmean+pvar,.];
"";
if fattail .ne 0;
"Conditional Degrees of Freedom";
bout[pmean+pvar+1:pmean+pvar+pfat,.];
"";endif;
if skew .ne 0;
"Conditional Skewness";
bout[pmean+pvar+pfat+1:pmean+pvar+pfat+1+skewn+skews,.];
"";endif;
"Log Likelihood " lik;
"";
"First four moments of standardized residuals";
mz~sez;
"";
"Kolmogorov-Smirnov " ks;
"Cramer-Von Mises   " vm;
"";
"Nyblom Statistics";
"  Joint  " nyblom;
if pmean .ne 0;
"  Mean   " nyblomm;
endif;
"  Var    " nyblomv;
if fattail == 1;
"  Tail   " nyblomf;
endif;
"";
output off;


title("CDF of Parametrically Transformed Residuals");
_pdate = "";
xtics(0,1,.2,2);
ytics(0,1,.2,2);
xy(zu,fzu~zu);

graphset;
_pdate="";
title("Density of Standardized Residual");
let _plegctl = 2 5 .5 5;
_plegstr = "Kernel\000Parametric";
kf = kernz~fz;
xy(br,kf);

title("Density of Residual");
kf = kerne~fe;
xtics(-5,5,1,2);
xy(br,kf);

xt = seqa(1973.5,1/52,t);
graphset;
fonts("complex");
_pltype = 6|4;
_pdate="";
xlabel("Observation");
m = y-e;


title("Series and Conditional Mean"); xy(xt,y~m);
@title("Residual"); xy(xt,e);@
title("Scaled Innovation"); xy(xt,z);
title("Conditional Variance"); xy(xt,h);
title("");

if fattail*(ta+ts) > 0;
brnu = seqa(2,.05,560);
kernu = nkern(nu,-.5,brnu,1);
ytics(0,20,5,2);
ylabel("Conditional Degrees of Freedom");
xy(xt,nu);
ytics(0,.5,.1,2);
ylabel("Inverse Degrees of Freedom");
xy(xt,1./nu);
graphset; _pdate="";
fonts("complex");
ylabel("Density");
xlabel("Degrees of Freedom");
xtics(0,20,5,2);
ytics(0,.15,.05,1);
xy(brnu,kernu);

endif;

if skew*(sa+ss) > 0;
graphset;
fonts("complex");
_pdate="";
xlabel("Observation");
ytics(-1,1,.4,2);
ylabel("Conditional Skewness Parameter"); xy(xt,lam);
brl = seqa(-1,.01,200);
kernl = nkern(lam,-.5,brl,1);
ylabel("Density");
xlabel("Skewness Parameter");
xtics(-1,1,.5,2);
ytics(0,7,1,2);
xy(brl,kernl);
endif;

/*
bt = b[pmean+pvar+1:pmean+pvar+pfat];
xx = seqa(-2,.1,50)';
yy = seqa(-1,.1,30);
zz = bt[1] + bt[2]*xx + bt[3]*(xx.^2);
zz = zz + yy*bt[4] + (yy.^2)*bt[5] + (yy*xx)*bt[6];
zz= 27.9/(1+exp(-zz)) + 2.1;
graphset;
xlabel("Lagged Innovation");
xtics(-2,3,1,2);
ylabel("Interest Differential");
ytics(-1,2,1,2);
_pframe = 0|0;
zlabel("Degrees of Freedom");
viewxyz(-10,-5,60);
surface(xx,yy,zz);
*/
